from dataclasses import dataclass
from typing import List, Optional
from unittest import TestCase

from pycarlo.features.metadata import (
    AllowBlockList,
    FilterEffectType,
    FilterType,
    MetadataFilter,
    MetadataFiltersContainer,
)


@dataclass
class DatasetEntry:
    project: str
    dataset: str


def _redshift_encoder(column: str, value: str, type: FilterType) -> str:
    if type == FilterType.REGEXP:
        op = "~"
        encoded_value = f"'^({value})$'"
    elif type == FilterType.PREFIX:
        op = " LIKE "
        encoded_value = f"'{value}%'"
    elif type == FilterType.SUFFIX:
        op = " LIKE "
        encoded_value = f"'%{value}'"
    elif type == FilterType.SUBSTRING:
        op = " LIKE "
        encoded_value = f"'%{value}%'"
    else:
        op = "="
        encoded_value = f"'{value}'"
    return f"{column}{op}{encoded_value}"


class TestDatasetFiltering(TestCase):
    DATASETS = [
        DatasetEntry("project_1", "dataset_1"),
        DatasetEntry("project_1", "dataset_2"),
        DatasetEntry("project_1", "dataset_3"),
        DatasetEntry("project_2", "dataset_1"),
        DatasetEntry("project_2", "dataset_2"),
        DatasetEntry("project_2", "dataset_3"),
        DatasetEntry("project_2", "dataset_4"),
    ]
    DATASETS_FOR_REGEXP = [
        DatasetEntry("project_1", "dataset_1"),
        DatasetEntry("project_1", "dataset_2"),
        DatasetEntry("project_2", "dataset_1"),
        DatasetEntry("project_2", "ds_2"),
        DatasetEntry("prj_3", "ds_1"),
        DatasetEntry("prj_3", "ds_2"),
        DatasetEntry("prj_3", "ds_3"),
        DatasetEntry("prj_3", "foobar_1"),
        DatasetEntry("prj_3", "foobar_3"),
    ]
    DATASETS_FOR_PREFIX = [
        DatasetEntry("project", "ds_1"),
        DatasetEntry("project", "ds_2"),
        *DATASETS_FOR_REGEXP,
    ]

    def test_block_single_dataset(self):
        # excludes only project_1.dataset_2
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_1", dataset="dataset_2", effect=FilterEffectType.BLOCK
                ),
            ]
        )
        self._validate_blocked_datasets(
            filters, [DatasetEntry(project="project_1", dataset="dataset_2")]
        )

    def test_block_multi_datasets(self):
        # excludes project_1.dataset_3 and project_2.dataset_1
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_1", dataset="dataset_3", effect=FilterEffectType.BLOCK
                ),
                MetadataFilter(
                    project="project_2", dataset="dataset_1", effect=FilterEffectType.BLOCK
                ),
            ]
        )
        limits = self._validate_blocked_datasets(
            filters,
            [
                DatasetEntry(project="project_1", dataset="dataset_3"),
                DatasetEntry(project="project_2", dataset="dataset_1"),
            ],
        )
        self.assertFalse(limits.is_whole_project_blocked("project_1"))
        self.assertFalse(limits.is_whole_project_blocked("project_2"))

    def test_block_single_project(self):
        # excludes all datasets in project_1
        filters = AllowBlockList(
            filters=[
                MetadataFilter(project="project_1", effect=FilterEffectType.BLOCK),
            ]
        )
        blocked = list(filter(lambda d: d.project == "project_1", self.DATASETS))
        limits = self._validate_blocked_datasets(filters, blocked)
        self.assertTrue(limits.is_whole_project_blocked("project_1"))
        self.assertFalse(limits.is_whole_project_blocked("project_2"))

    def test_block_second_project(self):
        # excludes all datasets in project_2
        filters = AllowBlockList(
            filters=[
                MetadataFilter(project="project_2", effect=FilterEffectType.BLOCK),
            ]
        )
        blocked = list(filter(lambda d: d.project == "project_2", self.DATASETS))
        limits = self._validate_blocked_datasets(filters, blocked)
        self.assertFalse(limits.is_whole_project_blocked("project_1"))
        self.assertTrue(limits.is_whole_project_blocked("project_2"))

    def test_block_single_project_regexp(self):
        # excludes all datasets in prj_3
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="prj_.*", type=FilterType.REGEXP, effect=FilterEffectType.BLOCK
                ),
            ]
        )
        blocked = list(filter(lambda d: d.project.startswith("prj_"), self.DATASETS_FOR_REGEXP))
        limits = self._validate_blocked_datasets(filters, blocked)
        self.assertFalse(limits.is_whole_project_blocked("project_1"))
        self.assertFalse(limits.is_whole_project_blocked("project_2"))
        self.assertTrue(limits.is_whole_project_blocked("prj_3"))

    def test_block_datasets_prefix(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project",
                    dataset="ds_",
                    type=FilterType.PREFIX,
                    effect=FilterEffectType.BLOCK,
                ),
                MetadataFilter(
                    project="prj_3",
                    dataset="ds_",
                    type=FilterType.PREFIX,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        self._validate_dataset_prefix(filters)

    def test_block_datasets_suffix(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2",
                    dataset="_2",
                    type=FilterType.SUFFIX,
                    effect=FilterEffectType.BLOCK,
                ),
                MetadataFilter(
                    project="prj_3",
                    dataset="_2",
                    type=FilterType.SUFFIX,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        self._validate_dataset_suffix(filters)

    def test_block_datasets_substring(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="prj_3",
                    dataset="bar",
                    type=FilterType.SUBSTRING,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        self._validate_dataset_substring(filters)

    def _validate_dataset_prefix(self, filters: AllowBlockList):
        blocked = [
            DatasetEntry("project", "ds_1"),
            DatasetEntry("project", "ds_2"),
            DatasetEntry("prj_3", "ds_1"),
            DatasetEntry("prj_3", "ds_2"),
            DatasetEntry("prj_3", "ds_3"),
        ]
        limits = self._validate_blocked_datasets(
            filters=filters, blocked_datasets=blocked, datasets=self.DATASETS_FOR_PREFIX
        )
        self.assertFalse(limits.is_whole_project_blocked("project_1"))
        self.assertFalse(limits.is_whole_project_blocked("project_2"))
        self.assertFalse(limits.is_whole_project_blocked("prj_3"))
        self.assertFalse(limits.is_whole_project_blocked("new_project"))

        self.assertTrue(limits.is_dataset_allowed("project_2", "dataset_1"))
        self.assertFalse(limits.is_dataset_allowed("project", "ds_1"))
        self.assertFalse(limits.is_dataset_allowed("project", "ds_2"))
        self.assertFalse(limits.is_dataset_allowed("prj_3", "ds_1"))
        self.assertFalse(limits.is_dataset_allowed("prj_3", "ds_2"))

    def _validate_dataset_suffix(self, filters: AllowBlockList):
        blocked = list(filter(lambda d: d.dataset.endswith("ds_2"), self.DATASETS_FOR_REGEXP))
        limits = self._validate_blocked_datasets(
            filters=filters, blocked_datasets=blocked, datasets=self.DATASETS_FOR_REGEXP
        )
        self.assertFalse(limits.is_whole_project_blocked("project_1"))
        self.assertFalse(limits.is_whole_project_blocked("project_2"))
        self.assertFalse(limits.is_whole_project_blocked("prj_3"))
        self.assertFalse(limits.is_whole_project_blocked("new_project"))

        self.assertTrue(limits.is_dataset_allowed("project_2", "dataset_1"))
        self.assertFalse(limits.is_dataset_allowed("project_2", "ds_2"))
        self.assertTrue(limits.is_dataset_allowed("prj_3", "ds_1"))
        self.assertFalse(limits.is_dataset_allowed("prj_3", "ds_2"))
        self.assertTrue(limits.is_dataset_allowed("prj_3", "ds_3"))

    def _validate_dataset_substring(self, filters: AllowBlockList):
        blocked = list(filter(lambda d: "bar" in d.dataset, self.DATASETS_FOR_REGEXP))
        limits = self._validate_blocked_datasets(
            filters=filters, blocked_datasets=blocked, datasets=self.DATASETS_FOR_REGEXP
        )

        self.assertFalse(limits.is_whole_project_blocked("project_1"))
        self.assertFalse(limits.is_whole_project_blocked("project_2"))
        self.assertFalse(limits.is_whole_project_blocked("prj_3"))
        self.assertFalse(limits.is_whole_project_blocked("new_project"))

        self.assertTrue(limits.is_dataset_allowed("project_2", "dataset_1"))
        self.assertTrue(limits.is_dataset_allowed("project_2", "ds_2"))
        self.assertTrue(limits.is_dataset_allowed("prj_3", "ds_1"))
        self.assertTrue(limits.is_dataset_allowed("prj_3", "ds_2"))
        self.assertTrue(limits.is_dataset_allowed("prj_3", "ds_3"))
        self.assertFalse(limits.is_dataset_allowed("prj_3", "foobar_1"))
        self.assertFalse(limits.is_dataset_allowed("prj_3", "foobar_3"))

    def test_no_filters(self):
        expected_datasets = self.DATASETS.copy()
        filtered_datasets = self._filter_datasets(self.DATASETS, MetadataFiltersContainer())
        self.assertEqual(expected_datasets, filtered_datasets)

    def test_allow_single_dataset(self):
        # excludes all datasets in project_1
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2", dataset="dataset_4", effect=FilterEffectType.ALLOW
                )
            ],
            default_effect=FilterEffectType.BLOCK,
        )
        limits = self._validate_allowed_datasets(
            filters, [DatasetEntry(project="project_2", dataset="dataset_4")]
        )

        self.assertTrue(limits.is_whole_project_blocked("project_1"))
        self.assertFalse(limits.is_whole_project_blocked("project_2"))

    def test_allow_second_project(self):
        # excludes all datasets in project_1
        filters = AllowBlockList(
            filters=[
                MetadataFilter(project="project_2", effect=FilterEffectType.ALLOW),
            ],
            default_effect=FilterEffectType.BLOCK,
        )
        allowed = list(filter(lambda d: d.project == "project_2", self.DATASETS))
        limits = self._validate_allowed_datasets(filters, allowed)

        self.assertTrue(limits.is_whole_project_blocked("project_1"))
        self.assertFalse(limits.is_whole_project_blocked("project_2"))

    def test_sql_query_block_single_project(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(project="project_2", effect=FilterEffectType.BLOCK),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_filtered)

        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual("(NOT(LOWER(database)='project_2'))", conditions)

    def test_sql_query_block_project_prefix(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="prj_", type=FilterType.PREFIX, effect=FilterEffectType.BLOCK
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_filtered)

        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual("(NOT(LOWER(database) LIKE 'prj_%'))", conditions)

    def test_sql_query_block_project_dataset_prefix(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_name",
                    dataset="ds_",
                    type=FilterType.PREFIX,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_filtered)

        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_name' AND LOWER(schema) LIKE 'ds_%'))", conditions
        )

    def test_sql_query_block_single_dataset(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2", dataset="dataset_1", effect=FilterEffectType.BLOCK
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_2' AND LOWER(schema)='dataset_1'))", conditions
        )

    def test_sql_query_block_single_project_case_sensitive(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="PROJECT_2", dataset="dataset_1", effect=FilterEffectType.BLOCK
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_2' AND LOWER(schema)='dataset_1'))", conditions
        )

    def test_sql_query_block_datasets_regexp(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2",
                    dataset="dataset_.+",
                    type=FilterType.REGEXP,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_2' AND LOWER(schema)~'^(dataset_.+)$'))", conditions
        )

    def test_sql_query_block_datasets_regexp_with_exception(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2",
                    dataset="dataset_.+",
                    type=FilterType.REGEXP,
                    effect=FilterEffectType.BLOCK,
                ),
                MetadataFilter(
                    project="project_2", dataset="dataset_10", effect=FilterEffectType.ALLOW
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "(((LOWER(database)='project_2' AND LOWER(schema)='dataset_10')) OR "
            "(NOT(LOWER(database)='project_2' AND LOWER(schema)~'^(dataset_.+)$')))",
            conditions,
        )

    def test_sql_query_project_block_single_dataset(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2", dataset="dataset_1", effect=FilterEffectType.BLOCK
                ),
                MetadataFilter(
                    project="project_1", dataset="dataset_2", effect=FilterEffectType.BLOCK
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            project="project_2",
            column_mapping={"project": "database", "dataset": "schema"},
            encoder=_redshift_encoder,
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_2' AND LOWER(schema)='dataset_1'))", conditions
        )

    def test_sql_query_project_block_nothing(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_1", dataset="dataset_2", effect=FilterEffectType.BLOCK
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertFalse(limits.is_project_with_datasets_filtered("project_2"))
        conditions = limits.get_sql_conditions(
            project="project_2",
            column_mapping={"project": "database", "dataset": "schema"},
            encoder=_redshift_encoder,
        )
        self.assertIsNone(conditions)

    def test_sql_query_block_multi_datasets(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2", dataset="dataset_1", effect=FilterEffectType.BLOCK
                ),
                MetadataFilter(
                    project="project_1", dataset="dataset_2", effect=FilterEffectType.BLOCK
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "("
            "NOT(LOWER(database)='project_2' AND LOWER(schema)='dataset_1') AND "
            "NOT(LOWER(database)='project_1' AND LOWER(schema)='dataset_2')"
            ")",
            conditions,
        )

    def test_sql_query_allow_multi_datasets(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2", dataset="dataset_1", effect=FilterEffectType.ALLOW
                ),
                MetadataFilter(
                    project="project_1", dataset="dataset_2", effect=FilterEffectType.ALLOW
                ),
            ],
            default_effect=FilterEffectType.BLOCK,
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "("
            "(LOWER(database)='project_2' AND LOWER(schema)='dataset_1') OR "
            "(LOWER(database)='project_1' AND LOWER(schema)='dataset_2')"
            ")",
            conditions,
        )

    def test_sql_query_block_case_insensitive(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="PROJECT_", effect=FilterEffectType.BLOCK, type=FilterType.PREFIX
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_filtered)

        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder, force_lowercase=False
        )
        self.assertEqual(
            "(NOT(database LIKE 'PROJECT_%'))",
            conditions,
        )

    def test_sql_query_block_datasets_prefix(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2",
                    dataset="dataset_",
                    type=FilterType.PREFIX,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_2' AND LOWER(schema) LIKE 'dataset_%'))", conditions
        )

    def test_sql_query_block_datasets_suffix(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2",
                    dataset="_2",
                    type=FilterType.SUFFIX,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_2' AND LOWER(schema) LIKE '%_2'))", conditions
        )

    def test_sql_query_block_datasets_suffix_without_project(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(dataset="_2", type=FilterType.SUFFIX, effect=FilterEffectType.BLOCK),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual("(NOT(LOWER(schema) LIKE '%_2'))", conditions)

    def test_sql_query_block_datasets_substring(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2",
                    dataset="s_2",
                    type=FilterType.SUBSTRING,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_2' AND LOWER(schema) LIKE '%s_2%'))", conditions
        )

    def test_sql_block_table_prefix(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2",
                    dataset="s_2",
                    table_name="tab",
                    type=FilterType.PREFIX,
                    effect=FilterEffectType.BLOCK,
                ),
            ]
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema", "table_name": "table_name"},
            _redshift_encoder,
        )
        self.assertEqual(
            "(NOT(LOWER(database)='project_2' AND LOWER(schema)='s_2' "
            "AND LOWER(table_name) LIKE 'tab%'))",
            conditions,
        )

    def test_whole_project_blocked(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2", dataset="dataset_1", effect=FilterEffectType.BLOCK
                ),
            ],
            default_effect=FilterEffectType.BLOCK,
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_blocked)
        self.assertTrue(limits.is_whole_project_blocked("project_2"))

        query = limits.get_sql_conditions(
            {"project": "database", "dataset": "schema"}, _redshift_encoder
        )
        self.assertTrue(query)

    def test_case_insensitive_comparison_exact_match(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_2", effect=FilterEffectType.BLOCK, type=FilterType.EXACT_MATCH
                ),
            ]
        )

        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_whole_project_blocked("PROJECT_2"))
        self.assertFalse(limits.is_whole_project_blocked("PROJECT_3"))

    def test_case_insensitive_comparison_prefix(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="project_", effect=FilterEffectType.BLOCK, type=FilterType.PREFIX
                ),
            ]
        )

        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_whole_project_blocked("PROJECT_1"))
        self.assertFalse(limits.is_whole_project_blocked("OTHER_PROJECT"))

    def test_case_insensitive_comparison_regex(self):
        filters = AllowBlockList(
            filters=[
                MetadataFilter(
                    project="[a-z]+_[1-9]+", effect=FilterEffectType.BLOCK, type=FilterType.REGEXP
                ),
            ]
        )

        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_whole_project_blocked("PROJECT_1"))
        self.assertFalse(limits.is_whole_project_blocked("PROJECT_"))

    def _validate_allowed_datasets(
        self, filters: AllowBlockList, expected_datasets: List, datasets: Optional[List] = None
    ) -> MetadataFiltersContainer:
        limits = MetadataFiltersContainer(metadata_filters=filters)
        filtered_datasets = self._filter_datasets(datasets or self.DATASETS, limits)
        self.assertEqual(expected_datasets, filtered_datasets)
        return limits

    def _validate_blocked_datasets(
        self, filters: AllowBlockList, blocked_datasets: List, datasets: Optional[List] = None
    ) -> MetadataFiltersContainer:
        expected_datasets = list(
            filter(lambda d: d not in blocked_datasets, (datasets or self.DATASETS))
        )
        limits = MetadataFiltersContainer(metadata_filters=filters)
        filtered_datasets = self._filter_datasets(datasets or self.DATASETS, limits)
        self.assertEqual(expected_datasets, filtered_datasets)
        return limits

    @staticmethod
    def _filter_datasets(
        datasets: List[DatasetEntry], limits: MetadataFiltersContainer
    ) -> List[DatasetEntry]:
        return [d for d in datasets if limits.is_dataset_allowed(d.project, d.dataset)]



import copy
import json
from typing import Dict, Optional
from unittest import TestCase
from unittest.mock import Mock, patch

from box import Box

from pycarlo.core import Client
from pycarlo.features.circuit_breakers import CircuitBreakerService
from pycarlo.features.circuit_breakers.exceptions import CircuitBreakerPollException
from pycarlo.features.exceptions import CircuitBreakerPipelineException

SAMPLE_BREACH_COUNT = 42
SAMPLE_RULE_ID = "1234"
SAMPLE_JOB_EXECUTION = "5678"
SAMPLE_TRIGGER = Box({"trigger_circuit_breaker_rule": {"job_execution_uuid": SAMPLE_JOB_EXECUTION}})
SAMPLE_TRIGGER_V2 = Box(
    {"trigger_circuit_breaker_rule_v2": {"job_execution_uuids": [SAMPLE_JOB_EXECUTION]}}
)
SAMPLE_ERROR = Box(
    {"payload": {"error": "Exception('this is a test')", "traceback": ["Line 10"]}},
    default_box_attr=None,
    default_box=True,
)


def sample_breach():
    return Box(
        {"payload": {"breach_count": SAMPLE_BREACH_COUNT}}, default_box_attr=None, default_box=True
    )


def sample_stage():
    return Box({"stage": "LOADING"}, default_box_attr=None, default_box=True)


def sample_payload():
    return Box(
        {"stage": "FREESTYLE", "payload": {"me": "silly"}}, default_box_attr=None, default_box=True
    )


class MockStateEmpty:
    get_circuit_breaker_rule_state_v2 = []


class MockState:
    get_circuit_breaker_rule_state_v2 = [
        Box(
            {"status": "PROCESSING_COMPLETE", "log": json.dumps([sample_stage(), sample_breach()])},
            default_box=True,
        )
    ]


class MockOutOfOrderState:
    get_circuit_breaker_rule_state_v2 = [
        Box(
            {
                "status": "PROCESSING_COMPLETE",
                "log": json.dumps([sample_payload(), sample_breach(), sample_stage()]),
            },
            default_box=True,
        )
    ]


class CircuitBreakerServiceTest(TestCase):
    def setUp(self) -> None:
        self._client_mock = Mock(spec=Client)

        self._service = CircuitBreakerService(mc_client=self._client_mock)

    @patch.object(CircuitBreakerService, "poll_all")
    def test_trigger_and_poll_in_breach(self, poll_mock: Mock):
        poll_mock.return_value = SAMPLE_BREACH_COUNT
        self.assertTrue(
            self._test_trigger_and_poll(poll_mock=poll_mock),  # type: ignore
        )

    @patch.object(CircuitBreakerService, "poll_all")
    def test_trigger_and_poll_no_breach(self, poll_mock: Mock):
        poll_mock.return_value = 0
        self.assertFalse(
            self._test_trigger_and_poll(poll_mock=poll_mock),  # type: ignore
        )

    @patch.object(CircuitBreakerService, "poll_all")
    def test_trigger_and_poll_by_rule_name_no_breach(self, poll_all_mock: Mock):
        poll_all_mock.return_value = 0
        self.assertFalse(
            self._test_trigger_and_poll_with_rule_name(poll_all_mock=poll_all_mock),  # type: ignore
        )

    @patch.object(CircuitBreakerService, "trigger_all")
    def _test_trigger_and_poll(self, trigger_all_mock: Mock, poll_mock: Mock) -> bool:
        trigger_all_mock.return_value = [SAMPLE_JOB_EXECUTION]

        in_breach = self._service.trigger_and_poll(rule_uuid=SAMPLE_RULE_ID)
        trigger_all_mock.assert_called_once_with(
            rule_uuid=SAMPLE_RULE_ID, namespace=None, rule_name=None, runtime_variables=None
        )
        poll_mock.assert_called_once_with(
            job_execution_uuids=[SAMPLE_JOB_EXECUTION], timeout_in_minutes=5
        )

        return bool(in_breach)

    @patch.object(CircuitBreakerService, "trigger_all")
    def _test_trigger_and_poll_with_rule_name(
        self,
        trigger_all_mock: Mock,
        poll_all_mock: Mock,
        runtime_variables: Optional[Dict[str, str]] = None,
    ) -> bool:
        trigger_all_mock.return_value = [SAMPLE_JOB_EXECUTION]

        in_breach = self._service.trigger_and_poll(
            rule_name="test_rule", runtime_variables=runtime_variables
        )
        trigger_all_mock.assert_called_once_with(
            rule_uuid=None,
            namespace=None,
            rule_name="test_rule",
            runtime_variables=runtime_variables,
        )
        poll_all_mock.assert_called_once_with(
            job_execution_uuids=[SAMPLE_JOB_EXECUTION], timeout_in_minutes=5
        )

        return bool(in_breach)

    def test_trigger(self):
        self._client_mock.return_value = SAMPLE_TRIGGER

        self.assertEqual(self._service.trigger(rule_uuid=SAMPLE_RULE_ID), SAMPLE_JOB_EXECUTION)
        self._client_mock.assert_called_once()
        self.assertEqual(
            self._trim_whitespace(self._client_mock.call_args[0][0]),
            self._trim_whitespace(
                f"mutation {{triggerCircuitBreakerRule(ruleUuid: "
                f'"{SAMPLE_RULE_ID}") {{jobExecutionUuid}}}}'
            ),
        )

    def test_trigger_all(self):
        self._client_mock.return_value = SAMPLE_TRIGGER_V2

        self.assertEqual(
            self._service.trigger_all(rule_uuid=SAMPLE_RULE_ID), [SAMPLE_JOB_EXECUTION]
        )
        self._client_mock.assert_called_once()
        self.assertEqual(
            self._trim_whitespace(self._client_mock.call_args[0][0]),
            self._trim_whitespace(
                f"mutation {{triggerCircuitBreakerRuleV2"
                f'(ruleUuid: "{SAMPLE_RULE_ID}") {{jobExecutionUuids}}}}'
            ),
        )

    def test_trigger_all_with_variables(self):
        self._client_mock.return_value = SAMPLE_TRIGGER_V2

        self.assertEqual(
            self._service.trigger_all(
                rule_uuid=SAMPLE_RULE_ID, runtime_variables={"var1": "val1", "var2": "val2"}
            ),
            [SAMPLE_JOB_EXECUTION],
        )
        self._client_mock.assert_called_once()
        self.assertEqual(
            self._trim_whitespace(self._client_mock.call_args[0][0]),
            self._trim_whitespace(
                f"mutation {{triggerCircuitBreakerRuleV2"
                f'(ruleUuid: "{SAMPLE_RULE_ID}",'
                'runtimeVariables:[{name:"var1",value:"val1"},{name:"var2",value:"val2"}]) '
                f"{{jobExecutionUuids}}}}"
            ),
        )

    @patch.object(CircuitBreakerService, "_poll")
    def test_poll_in_breach(self, internal_poll_mock: Mock):
        internal_poll_mock.return_value = [sample_breach()]
        self.assertEqual(
            self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION), SAMPLE_BREACH_COUNT
        )
        internal_poll_mock.assert_called_once_with(
            job_execution_uuids=[SAMPLE_JOB_EXECUTION], timeout_in_minutes=5
        )

    @patch.object(CircuitBreakerService, "_poll")
    def test_poll_not_in_breach(self, internal_poll_mock: Mock):
        log = copy.deepcopy(sample_breach())
        log.payload.breach_count = 0

        internal_poll_mock.return_value = [log]
        self.assertEqual(self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION), 0)
        internal_poll_mock.assert_called_once_with(
            job_execution_uuids=[SAMPLE_JOB_EXECUTION], timeout_in_minutes=5
        )

    @patch.object(CircuitBreakerService, "_poll")
    def test_poll_with_pipeline_error(self, internal_poll_mock: Mock):
        internal_poll_mock.return_value = [SAMPLE_ERROR]
        with self.assertRaises(CircuitBreakerPipelineException) as context:
            self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION)
        self.assertEqual(
            str(context.exception), f"Execution pipeline errored out. Details:\n{SAMPLE_ERROR}"
        )

    @patch.object(CircuitBreakerService, "_poll")
    def test_poll_with_timeout_and_empty_response(self, internal_poll_mock: Mock):
        internal_poll_mock.return_value = []
        with self.assertRaises(CircuitBreakerPollException) as context:
            self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION)
        self.assertEqual(str(context.exception), "Polling timed out or contains a malformed log.")

    @patch.object(CircuitBreakerService, "_poll")
    def test_poll_with_timeout_and_none_response(self, internal_poll_mock: Mock):
        internal_poll_mock.return_value = None
        with self.assertRaises(CircuitBreakerPollException) as context:
            self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION)
        self.assertEqual(str(context.exception), "Polling timed out or contains a malformed log.")

    def test_internal_poll(self):
        self._client_mock.return_value = MockState
        self.assertEqual(
            [sample_breach()],
            self._service._poll(job_execution_uuids=[SAMPLE_JOB_EXECUTION], timeout_in_minutes=10),
        )

    def test_internal_poll_out_of_order_logs(self):
        self._client_mock.return_value = MockOutOfOrderState
        self.assertEqual(
            [sample_breach()],
            self._service._poll(job_execution_uuids=[SAMPLE_JOB_EXECUTION], timeout_in_minutes=10),
        )

    @patch("pycarlo.features.circuit_breakers.service.time")
    def test_internal_poll_with_timeout(self, mock_time: Mock):
        self._client_mock.return_value = MockStateEmpty
        mock_time.time.side_effect = [1643751961, 1643752021, 1644356761]
        self.assertIsNone(
            self._service._poll(job_execution_uuids=[SAMPLE_JOB_EXECUTION], timeout_in_minutes=10)
        )
        mock_time.sleep.assert_called_once_with(15)

    @staticmethod
    def _trim_whitespace(_str: str) -> str:
        return "".join(str(_str).split())
